from contextlib import suppress
from typing import Any, Dict, List

from selenium import webdriver
from selenium.common.exceptions import (
    InvalidArgumentException,
    TimeoutException,
    UnexpectedAlertPresentException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


class Crawler:
    def __init__(
        self,
        headless=True,
        crawler_wait_seconds=30,
        page_load_timeout=15,
        firefox_profile_preferences: Dict[str, Any] = None,
        firefox_options: List[str] = None,
        driver_params: Dict[str, Any] = None,
    ):
        """
        Args:
            headless (bool): Whether to run Firefox in headless mode. Defaults to True.
            crawler_wait_seconds (int): Maximum time to wait for elements to become present. Defaults to 30.
            page_load_timeout (int): Maximum time to wait for the page to load. Defaults to 15.
            firefox_profile_preferences (Dict[str, Any]): Dictionary of Firefox profile preferences. Defaults to None.
            firefox_options (List[str]): List of command-line options for Firefox. Defaults to None.
            driver_params (Dict[str, Any]): Additional parameters to pass to the Firefox driver. Defaults to None.
        """
        self.headless = headless
        self.crawler_wait_seconds = crawler_wait_seconds
        self.page_load_timeout = page_load_timeout
        self._firefox_profile_preferences = firefox_profile_preferences
        self._firefox_options = firefox_options
        self._driver_params = driver_params if driver_params else {}
        self._driver = None

    @property
    def firefox_profile(self):
        """
        Creates and configures a Firefox profile with the specified preferences.

        Returns:
            webdriver.FirefoxProfile: The configured Firefox profile.
        """
        profile = webdriver.FirefoxProfile()
        if self._firefox_profile_preferences:
            for preference_name, preference_value in self._firefox_profile_preferences.items():
                profile.set_preference(preference_name, preference_value)
        return profile

    @property
    def firefox_options(self):
        """
        Creates and configures Firefox options, including headless mode and additional arguments.

        Returns:
            Options: The configured Firefox options.
        """
        firefox_options = Options()
        if self.headless:
            firefox_options.add_argument("--headless")
        if self._firefox_options:
            for option in self._firefox_options:
                firefox_options.add_argument(option)
        firefox_options.profile = self.firefox_profile
        return firefox_options

    @property
    def driver(self):
        """
        Creates and returns the Firefox WebDriver instance.  Creates it only if it doesn't already exist.

        Returns:
            webdriver.Firefox: The Firefox WebDriver instance.
        """
        if not self._driver:
            self._driver = webdriver.Firefox(options=self.firefox_options, **self._driver_params)
            self._driver.set_page_load_timeout(self.page_load_timeout)
        return self._driver

    def open_url(self, url, tries=3):
        """
        Opens the specified URL in the browser. Retries the operation if it fails due to timeout or alert.

        Args:
            url (str): The URL to open.
            tries (int): The number of times to retry the operation. Defaults to 3.

        Raises:
            Exception: If the URL cannot be opened after the specified number of retries.
        """
        last_exception: BaseException | None = None
        while tries > 0:
            try:
                self.driver.get(url)
                return
            except (TimeoutException, InvalidArgumentException, UnexpectedAlertPresentException) as e:
                tries -= 1
                last_exception = e
        raise last_exception

    def wget(
        self, selector, driver=None, single=False, wait=True, wait_seconds: float | None = None
    ) -> WebElement | List[WebElement]:
        """
        Fetches web elements matching a CSS selector from a webpage.

        Args:
            selector (str): The CSS selector to locate the desired elements.
            driver (webdriver or WebElement, optional): The Selenium WebDriver instance or a WebElement to search within. Defaults to the crawler's driver.
            single (bool, optional): If True, returns only the first element found. If False, returns all matching elements. Defaults to False.
            wait (bool, optional): If True, waits for the elements to be present on the page before returning them. Defaults to True.
            wait_seconds (float, optional): The maximum time to wait for the elements to become present, in seconds. Defaults to the crawler's default wait time.

        Returns:
            list: A list of WebElement objects matching the selector. If `single` is True and an element is found, a list containing only that element is returned. Returns an empty list if no elements are found.

        Raises:
            TimeoutException: If `wait` is True and the elements do not become present within the specified `wait_seconds`.
        """
        wait_seconds = wait_seconds if wait_seconds is not None else self.crawler_wait_seconds

        driver = driver if driver is not None else self.driver

        wait_func = EC.presence_of_all_elements_located if not single else EC.presence_of_element_located
        if wait:
            result = WebDriverWait(driver, wait_seconds).until(wait_func((By.CSS_SELECTOR, selector)))
        else:
            result = driver.find_elements_by_css_selector(selector)
        return result

    def wgetx(
        self, xpath, driver=None, single=False, wait=True, wait_seconds: float | None = None
    ) -> WebElement | List[WebElement]:
        """
        Fetches web elements matching an XPath expression from a webpage.

        Args:
            xpath (str): The XPath expression to locate the desired elements.
            driver (webdriver or WebElement, optional): The Selenium WebDriver instance or a WebElement to search within. Defaults to the crawler's driver.
            single (bool, optional): If True, returns only the first element found. If False, returns all matching elements. Defaults to False.
            wait (bool, optional): If True, waits for the elements to be present on the page before returning them. Defaults to True.
            wait_seconds (float, optional): The maximum time to wait for the elements to become present, in seconds. Defaults to the crawler's default wait time.

        Returns:
            list: A list of WebElement objects matching the XPath expression. If `single` is True and an element is found, a list containing only that element is returned. Returns an empty list if no elements are found.

        Raises:
            TimeoutException: If `wait` is True and the elements do not become present within the specified `wait_seconds`.
        """
        wait_seconds = wait_seconds if wait_seconds is not None else self.crawler_wait_seconds
        driver = driver if driver is not None else self.driver

        wait_func = EC.presence_of_all_elements_located if not single else EC.presence_of_element_located
        if wait:
            result = WebDriverWait(driver, wait_seconds).until(wait_func((By.XPATH, xpath)))
        else:
            result = driver.find_elements_by_xpath(xpath)
        return result

    def click_button(self, css_selector: str):
        """
        Clicks a button located by its CSS selector.

        Args:
            css_selector (str): The CSS selector to identify the button.

        Returns:
            None

        Raises:
            NoSuchElementException: If no button is found matching the CSS selector.
            WebDriverException: If an error occurs while clicking the button.
        """
        button: WebElement = self.wget(css_selector, single=True)
        self.click(button)

    def click_buttonx(self, xpath: str):
        """
        Clicks a button located by its XPath.

        Args:
            xpath (str): The XPath to identify the button.

        Returns:
            None

        Raises:
            NoSuchElementException: If no button is found matching the XPath.
            WebDriverException: If an error occurs while clicking the button.
        """
        button: WebElement = self.wgetx(xpath, single=True)
        self.click(button)

    def reset(self):
        """
        Quits the WebDriver instance and resets the internal driver variable.
        """
        if self._driver:
            with suppress(BaseException):
                self._driver.quit()
            self._driver = None

    def click(self, element: WebElement):
        """
        Clicks on the given WebElement using JavaScript.

        Args:
            element (WebElement): The element to click.
        """
        self.driver.execute_script("arguments[0].click();", element)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reset()

    def __del__(self):
        self.reset()
